/*
 * fas.c  -  process FAS host adapters.
 *
 * Author: John DiMarco, University of Toronto, CDF
 *         jdd@cdf.toronto.edu
 * 
 */

#ifndef lint
static char rcsid[] = "$Id: fas.c,v 1.8 1999/04/22 17:40:16 jdd Exp $";
#endif

#ifdef FAS_SUPPORT

#include "uscsi.h"
#include "format.h"
#include <stdio.h>

#ifdef FILENAME_MAX     /* defined in stdio.h only for SunOS >=5.x */
#define SYSV
#endif

#include <sys/param.h>
#include <values.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <nlist.h>
#include <fcntl.h>
#include <sys/utsname.h>
#include "utils.h"
#include "common.h"
#include "param.h"

#ifdef SYSV
#include "sunos5.h"
#include <sys/scsi/adapters/fasreg.h>
#include <sys/scsi/adapters/fasvar.h>

#undef SYNC_KNOWN
#undef NODISC
#undef NOTAG
#undef NOWIDE
#undef NOISY

#ifndef SYNC_KNOWN
#define SYNC_KNOWN(i) (fas->f_sync_known & (1<<(i)))
#endif /* SYNC_KNOWN */

#ifdef notdef	/* f_nodisc doesn't exit */
#ifndef NODISC
#define NODISC(i) (fas->f_nodisc & (1<<(i)))
#endif /* NODISC */
#endif /* notdef */

#ifndef NOTAG
#define NOTAG(i) (fas->f_notag & (1<<(i)))
#endif /* NOTAG */

#ifndef NOWIDE
#define NOWIDE(i) (fas->f_nowide & (1<<(i)))
#endif /* NOWIDE */

#ifndef NOISY
#define NOISY(i) (fas->f_backoff & (1<<(i)))
#endif /* NOISY */

/* 
 * The following is a sanity check. It makes no sense if the same device
 * is both sync and nosync, or both wide and nowide.
 */
#ifndef INSANE
#define INSANE ((fas->f_sync_enabled & fas->f_nosync) || (fas->f_wide_enabled & fas->f_nowide))
#endif /* INSANE */

static char *fasType(typecode)
unsigned int typecode;
{
	/* see /usr/include/sys/scsi/adapters/fasreg.h */
	switch(typecode){
	case 0:		return("FAS366");
	case 5:		return("FAST");
	default:	return("unknown");
	}
}

static u_int fas_get_speed(fas, i)
struct fas *fas;
int i; /* target index */
{
	u_int cycletime, speed;

	/*
	 * Synch cycle time in milliseconds is the
	 * number of ticks in the period divided
	 * by the tick rate. fas_clock_conv is the
	 * fas's tick rate in units of 5MHz.
	 */
	cycletime = fas->f_clock_conv ? (fas->f_sync_period[i] *
			(1000/(int)(5*fas->f_clock_conv))) : 0;

	/*
	 * Compute speed in thousanths of MBs/sec.
	 */
	speed = cycletime?((((1000*1000*1000)/cycletime)+999)/1000):0;

	if(!NOWIDE(i)) speed *= 2;

#ifdef DEBUG
        dfprintf(2, stderr, "Target %d: clock %u, minperiod %u, nowide %u\n",
		i, fas->f_clock_conv, fas->f_sync_period[i], NOWIDE(i));
#endif /* DEBUG */

	return(speed);
}

static void fas_controller_message(cnum, fas)
int cnum;
struct fas *fas;
{
	printf("fas%d:", cnum);
	printf(" type %d (%s)", (u_int)fas->f_type,
		fasType(fas->f_type));
	printf(" speed %dMHz", (u_int)fas->f_clock_conv*5);
	printf("\n");
}

static void fas_synch_message(fas, i, speed)
struct fas *fas;
int i;
u_int speed;
{
	if(speed && SYNC_KNOWN(i)){
		printf("Synchronous(%d.%.3dMB/sec)", speed/1000, speed%1000);
	} else {
		printf("Asynchronous");
	}
	if(NOISY(i)){
		printf(" Noisy");
	} else {
		printf(" Clean");
	}
#ifdef notdef
	if(NODISC(i)){
		printf(" CannotReconnect");
	} else {
		printf(" CanReconnect");
	}
#endif /* notdef */
	if(NOTAG(i)){
		printf(" NoTaggedQueuing");
	} else {
		printf(" TaggedQueuing");
	}
	if(NOWIDE(i)){
		printf(" Narrow");
	} else {
		printf(" Wide");
	}
}

int process_fas()
{
	u_long addr;
	struct fas fas[1];
	int noisycount=0; /* number of noisy devices */
	u_int j, nfas;

        if(-1==kvm(nl[N_FAS_HEAD].n_value, addr)){
		return(0); /* fas driver not loaded */
        }
        if((u_long)NULL==addr) return(0); /* no fas's */

        if(-1==kvm(addr, fas[0])){
                Error("kvm read: initial fas");
        }

	nfas=get_nfas();
	if(0==nfas) return(0);

	for(j=0;j<nfas;j++){
		struct dev_info di, *next;
		char *devicepath, *dname;

		if((u_long)NULL==addr) break;

                if(-1==kvm(addr, fas[0])){
			Error("kvm read: fas #%d", j);
		}

		/* 
		 * grab the devinfo structure for the fas controller. CNUM 
		 * requires it, and we need it to determine the device names.
		 */
		if(-1==kvm(fas->f_dev, devi)){
			Error("kvm read: #%d fas devinfo", j);
		} 

		devicepath = get_devicepath(j);

		fas->f_dev=(dev_info_t)&devi; 

#ifdef INSANE
                if(INSANE){
                        Warning(
"The kernel data structure for fas%d failed a sanity check.\n\
  There may be a mismatch between the compile-time headers and the currently\n\
  running kernel.  A recompile on this machine may be necessary.", CNUM);
                }
#endif /* INSANE */


		if(param_is_set(P_CONTROLLER)) {
			fas_controller_message(CNUM, fas);
		}

		init_scsitarget();

		/* 
		 * Now, traverse fas's devinfo child list, looking for
		 * attached drivers (i.e. DDI_CF2 is true). If we find
		 * an attached driver, we presume it is an active device.
		 */

		for(next=ddi_get_child(&devi);
		    NULL!=next; 
		    next=ddi_get_next_sibling(&di)){
			char *devicename, *addrname;
			int target, lun;

			if(-1==kvm(next, di)){
			    Error("kvm read: #%d fas devinfo child", j);
			}

			if(!DDI_CF2(&di)) continue; /* no device attached */

			devicecount++;

			/*
			 * Get the device name and address, and 
			 * compute the device path for it
			 */

			/* get device name */
			dname=s(kvmstr((u_long)ddi_get_name(&di)));
			/* ... and address */
			addrname=s(get_addr(&di));

			devicename = compute_devicename(dname, 
						devicepath, addrname);

			if(-1==get_scsiaddr(dname, &di, &target, &lun)){
				Error("scsiaddr: #%d", j);
			}

			if(0==lun && NOISY(target)) noisycount++;

			banner(param_is_set(P_FORMATDAT), 
				param_is_set(P_REPORT), devicecount);
			if(!param_is_set(P_QUIET)){
				if(param_is_set(P_SUNOS4STYLE)) {
					if(lun==0){
						boot_synch_message("fas",
							CNUM, target, 
							fas_get_speed(fas, 
								target), 
								NARROW);
					}
				} else {
					printf("fas%d: %s%s tgt %d lun %d:\n", 
						CNUM, dname, addrname, 
						target, lun);
					if(param_is_set(P_SYNCH)){
						printf("\t");
						fas_synch_message(fas, target, 
						 fas_get_speed(fas, target));
						printf("\n");
					}
				}
			}
			if(param_is_set(P_PROBE)) {
				do_scsi_probe(dname, devicename, lun);
			}
			if(param_is_set(P_REPORT)) {
				report(param_is_set(P_FORMATDAT),
					osrelease, devicename,
					param_is_set(P_ORIGINAL));
			}
			free(devicename);
			free(addrname);
			free(dname);
		}

		free(devicepath);

                addr=(u_long)fas->f_next;

	}
	return(noisycount);
}

#else /* !SYSV */
/* FAS isn't supported on SunOS 4.x */
int process_fas(){}
#endif /* SYSV */
#endif /* FAS_SUPPORT */
