/*
 * ptisp.c  -  process PTISP host adapters.
 *
 * Author: Peter Vernam, Draper Laboratory <vernam@draper.com>
 * 
 * This module was adapted from John DiMarco's (University of Toronto, CDF
 * <jdd@cdf.toronto.edu>) esp.c and isp.c modules to provide support for the
 * Performance Technologies (315 Science Parkway, Rochester, NY 14620;
 * 716-256-0248) ptisp driver and host adapter.
 */

#ifndef lint
static char rcsid[] = "$Id: ptisp.c,v 1.10 1999/04/22 17:40:16 jdd Exp $";
#endif

#ifdef PTISP_SUPPORT

#include "uscsi.h"
#include "format.h"
#include <stdio.h>

#ifdef FILENAME_MAX	/* defined in stdio.h only for SunOS >=5.x */
#define SYSV
#endif

#include <sys/param.h>
#include <values.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <nlist.h>
#include <fcntl.h>
#include <sys/utsname.h>
#include "utils.h"
#include "common.h"
#include "param.h"

typedef struct {
	int     period;         /* ISP_xM_SYNC_PERIOD */
	int     speed;          /* x000 */
} isp_speed_mapping;


#ifdef SYSV
#include "sunos5.h"
#ifndef VERS_ACCHDL
	/* 
	 * ddi_acc_handle_t is used in ispvar5.h, but not defined
	 * under Solaris 2.4. Since the include file in 2.5 that defines
	 * it also defines VERS_ACCHDL, if VERS_ACCHDL isn't defined, we'd
	 * better define it before including ispvar5.h.
	 */
typedef void *ddi_acc_handle_t;
#endif /* VERS_ACCHDL */

#include "ptisp/ispmail5.h"
#include "ptisp/ispvar5.h"

#define TEN_M ISP_10M_SYNC_PERIOD
#define EIGHT_M ISP_8M_SYNC_PERIOD
#define FIVE_M ISP_5M_SYNC_PERIOD
#define FOUR_M ISP_4M_SYNC_PERIOD

#define PTISP_SYNC isp_synch
#define PTISP_CAP isp_cap
#define PTISP_MAJOR_REV isp_major_rev
#define PTISP_MINOR_REV isp_minor_rev
#define PTISP_DIFFMODE ptisp_diffmode
#define PTISP_DEV isp_dip

#define N_PTISP_SYMBOL N_PTISP_HEAD

/* 
 * It's ugly and dangerous to redefine CNUM here, but SunOS 5.4 ispvar.h's 
 * idea of CNUM is completely wrong. It uses data structures that existed in
 * 2.3 but no longer exist in 2.4.
 */
#undef CNUM
#define CNUM		(ddi_get_instance(ptisp->isp_tran->tran_hba_dip))

#else /* !SYSV */
#include <scsi/scsi_types.h>
#include <scsi/impl/transport.h>
#include "sunos4.h"
#include "ptisp/ispmail4.h"
#include "ptisp/ptisp4.h"

#define TEN_M ISP_CAP_10M_SYNC_PERIOD
#define EIGHT_M ISP_CAP_8M_SYNC_PERIOD
#define FIVE_M ISP_CAP_5M_SYNC_PERIOD
#define FOUR_M ISP_CAP_4M_SYNC_PERIOD

#define PTISP_SYNC sync
#define PTISP_CAP cap
#define PTISP_MAJOR_REV major_rev
#define PTISP_MINOR_REV minor_rev
#define PTISP_DIFFMODE diffmode
#define PTISP_DEV dev

#define N_PTISP_SYMBOL N_PTISP_SOFTC

#define PERIOD_MASK(val)	((val) & 0xff)
#undef CNUM
#define CNUM			(ptisp->unit_no)

#endif /* !SYSV */

static isp_speed_mapping isp_speed[] = {
#ifdef USE_SUN_ISP_SPEEDS
	TEN_M, 		10000,
	EIGHT_M,   	8000,
	FIVE_M,   	5000,
	FOUR_M,   	4000
#define ISP_PERIOD_TABLE 4      /* Number of elements in isp_speed[] */
#else /* !USE_SUN_ISP_SPEEDS */
	0x0c, 		20000,
	0x10, 		15000,
	0x15, 		12000,
	0x19, 		10000,
	0x1e,  		8000,
	0x21,  		7000,
	0x25,  		6500,
	0x2a,  		6000,
	0x2d,  		5500,
	0x32,  		5000,
	0x38,  		4500,
	0x3a,  		4000,
	0x42,  		3000
#define ISP_PERIOD_TABLE 13     /* Number of elements in isp_speed[] */
#endif /* !USE_SUN_ISP_SPEEDS */
};


static u_int ptisp_get_speed(ptisp, i)
struct ptisp *ptisp;
int i; /* target index */
{
	int index;

	for (index = 0; index < ISP_PERIOD_TABLE; index++){
		if(isp_speed[index].period==PERIOD_MASK(ptisp->PTISP_SYNC[i])){
			return(isp_speed[index].speed);
		}
	}
	return((u_int)-1);
}

static void ptisp_synch_message(ptisp, i, speed)
struct ptisp *ptisp;
int i;
u_int speed;
{
	int width = NARROW;
	u_short ptisp_cap = ptisp->PTISP_CAP[i];
	if(ptisp_cap & ISP_CAP_WIDE) width = WIDE;

	switch (speed) {
		case (u_int)-1 :
			if ( ptisp_cap & ISP_CAP_SYNC ) {
				printf("Synchronous(Unknown MB/sec)" );
			}
			else {
				printf("Asynchronous");
			}
			break;
		case 0 :
			printf("Asynchronous");
			break;
		default :
			if(speed && (ptisp_cap & ISP_CAP_SYNC)){
				printf("Synchronous(%d.%.3dMB/sec)",
					(width/8)*speed/1000, 
					(width/8)*speed%1000);
			}
			else {
				printf("Asynchronous"); 
			}
			break;
	}
	printf(" Unknown"); /* ptisp driver doesn't keep track of errors */
	if(ptisp_cap & ISP_CAP_DISCONNECT){
		printf(" CanReconnect");
	} else {
		printf(" CannotReconnect");
	}
	if(ptisp_cap & ISP_CAP_TAG){
		printf(" TaggedQueuing");
	} else {
		printf(" NoTaggedQueuing");
	}
	if(width == WIDE){
		printf(" Wide");
	} else {
		printf(" Narrow");
	}
}

static void ptisp_controller_message(cnum, ptisp)
int cnum;
struct ptisp *ptisp;
{
	printf("ptisp%d:", cnum);
	printf(" PTI/QLGC ISP 1000" );
	printf(" firmware revision %d.%d", (int) ptisp->PTISP_MAJOR_REV,
		(int) ptisp->PTISP_MINOR_REV );
	if(ptisp->PTISP_DIFFMODE) {
		printf(", Differential");
	} else {
		printf(", Single-ended");
	}
	printf("\n");
}

int process_ptisp()
{
	u_long addr;
	struct ptisp ptisp[1];
	u_int j, nptisp;
	int noisycount=0;
#ifndef SYSV
	struct ptisp *ptisp_softc, *ptisp_p;
#endif /* !SYSV */
	if (!nl[N_PTISP_SYMBOL].n_type) {
		return(0); /* ptisp driver not loaded */
	}

	if(-1==kvm(nl[N_PTISP_SYMBOL].n_value, addr)){
		return(0); /* ptisp driver not loaded */
	}

	if((u_long)NULL==addr) return(0); /* no ptisp's */

        if(-1==kvm(addr, ptisp[0])){
		Error("kvm read: initial ptisp");
	}

#ifndef SYSV
	ptisp_softc = ptisp_p = (struct ptisp *)addr;
#endif /* !SYSV */

	nptisp=get_nptisp();
	if(0==nptisp) return(0);

	for(j=0;j<nptisp;j++){
		struct dev_info di, *next;
		char *devicepath, *dname;

		if((u_long)NULL==addr) break;

                if(-1==kvm(addr, ptisp[0])){
			Error("kvm read: ptisp #%d", j);
		}

#ifdef SCSI_HBA_TRAN_CLONE
		{
			scsi_hba_tran_t tran;
			/* 
			 * Grab the transport devinfo structure for the ptisp 
			 * controller. Needed for CNUM. 
			 */
			if(-1==kvm((u_long)ptisp->isp_tran, tran)){
				Error("kvm read: #%d ptisp tran", j);
			}
			ptisp->isp_tran=(scsi_hba_tran_t *)&tran; 
			if(-1==kvm((u_long)ptisp->isp_tran->tran_hba_dip, 
					devi)){
				Error("kvm read: #%d ptisp devinfo", j);
			}
			ptisp->isp_tran->tran_hba_dip=(dev_info_t *)&devi;
		}
#else /* !SCSI_HBA_TRAN_CLONE */
		/* 
		 * grab the devinfo structure for the ptisp controller. CNUM 
		 * requires it, and we need it to determine the device names.
		 */
		if(-1==kvm((u_long)ptisp->PTISP_DEV, devi)){
			Error("kvm read: #%d ptisp devinfo", j);
		}
		ptisp->PTISP_DEV=(dev_info_t)&devi; /* for CNUM */
#endif /* !SCSI_HBA_TRAN_CLONE */

		devicepath = get_devicepath(j);

		if(param_is_set(P_CONTROLLER)) {
			ptisp_controller_message(CNUM, ptisp);
		}

		init_scsitarget();

		/* 
		 * Now, traverse ptisp's devinfo child list, looking for
		 * attached drivers (i.e. DDI_CF2 is true). If we find
		 * an attached driver, we presume it is an active device.
		 */

		for(next=ddi_get_child(&devi);
		    NULL!=next; 
		    next=ddi_get_next_sibling(&di)){
			char *devicename, *addrname;
			int target, lun;

			if(-1==kvm((u_long)next, di)){
			    Error("kvm read: #%d ptisp devinfo child", j);
			}

			if(!DDI_CF2(&di)) continue; /* no device attached */

			devicecount++;

			/*
			 * Get the device name and address, and 
			 * compute the device path for it
			 */

			/* get device name */
			dname=s(kvmstr((u_long)ddi_get_name(&di)));
			/* ... and address */
			addrname=s(get_addr(&di));

#ifdef CENTRAL_DATA_CT
			if(STREQ(CENTRAL_DATA_CT, dname)) {
				/* special case for Central Data STS driver */
				if(!di.devi_minor) continue;
			} 
#endif /* CENTRAL_DATA_CT */

			devicename = compute_devicename(dname, 
						devicepath, addrname);

			if(-1==get_scsiaddr(dname, &di, &target, &lun)){
				Error("scsiaddr: #%d", j);
			}

			banner(param_is_set(P_FORMATDAT), 
				param_is_set(P_REPORT), devicecount);
			if(!param_is_set(P_QUIET)){
				if(param_is_set(P_SUNOS4STYLE)) {
					if(lun==0){
						boot_synch_message("ptisp",
							CNUM, target, 
						  	ptisp_get_speed(ptisp,
								target),
								WIDE);
					}
				} else {
					printf(
					     "ptisp%d: %s%s tgt %d lun %d:\n",
						CNUM, dname, addrname, 
						target, lun);
					if(param_is_set(P_SYNCH)){
					    printf("\t");
					    ptisp_synch_message(ptisp, target, 
					     ptisp_get_speed(ptisp, target));
					    printf("\n");
					}
				}
			}


			if(param_is_set(P_REPORT)){
				report(param_is_set(P_FORMATDAT), 
					osrelease, devicename,
					param_is_set(P_ORIGINAL));
			}
			if(param_is_set(P_PROBE)){
				do_scsi_probe(dname, devicename, lun);
			}
			free(devicename);
			free(addrname);
			free(dname);
		}
		free(devicepath);
#ifdef SYSV
                addr=(u_long)ptisp->isp_next;
#else
                addr=(u_long)(++ptisp_p);
#endif
	}

	return(noisycount);

}

#endif /* PTISP_SUPPORT */
